﻿#----------------------------------------------------------#
#- MusicBeeIPCSDK Rb v2.0.0                               -#
#- Copyright © Kerli Low 2014                             -#
#- This file is licensed under the                        -#
#- BSD 2-Clause License                                   -#
#- See LICENSE_MusicBeeIPCSDK for more information.       -#
#----------------------------------------------------------#

require_relative "enums.rb"
require_relative "constants.rb"
require_relative "pack.rb"
require_relative "unpack.rb"
require_relative "winapi.rb"


class MusicBeeIPC
  # @!group General

  # @return [bool]
  def probe()
    send_message(find_hwnd(), WM_USER, MBC_Probe, 0) != MBE_Error
  end

  # @return [MBError]
  def play_pause()
    send_message(find_hwnd(), WM_USER, MBC_PlayPause, 0)
  end

  # @return [MBError]
  def play()
    send_message(find_hwnd(), WM_USER, MBC_Play, 0)
  end

  # @return [MBError]
  def pause()
    send_message(find_hwnd(), WM_USER, MBC_Pause, 0)
  end

  # @return [MBError]
  def stop()
    send_message(find_hwnd(), WM_USER, MBC_Stop, 0)
  end

  # @return [MBError]
  def stop_after_current()
    send_message(find_hwnd(), WM_USER, MBC_StopAfterCurrent, 0)
  end

  # @return [MBError]
  def previous_track()
    send_message(find_hwnd(), WM_USER, MBC_PreviousTrack, 0)
  end

  # @return [MBError]
  def next_track()
    send_message(find_hwnd(), WM_USER, MBC_NextTrack, 0)
  end

  # @return [MBError]
  def start_auto_dj()
    send_message(find_hwnd(), WM_USER, MBC_StartAutoDj, 0)
  end

  # @return [MBError]
  def end_auto_dj()
    send_message(find_hwnd(), WM_USER, MBC_EndAutoDj, 0)
  end

  # @return [MBPlayState]
  def get_play_state()
    send_message(find_hwnd(), WM_USER, MBC_GetPlayState, 0)
  end

  # @return [String]
  def get_play_state_str()
    case send_message(find_hwnd(), WM_USER, MBC_GetPlayState, 0)
    when MBPS_Loading
      "Loading"
    when MBPS_Playing
      "Playing"
    when MBPS_Paused
      "Paused"
    when MBPS_Stopped
      "Stopped"
    else
      "Undefined"
    end
  end

  # @return [String]
  def get_play_state_sym()
    case send_message(find_hwnd(), WM_USER, MBC_GetPlayState, 0)
    when MBPS_Loading
      :Loading
    when MBPS_Playing
      :Playing
    when MBPS_Paused
      :Paused
    when MBPS_Stopped
      :Stopped
    else
      :Undefined
    end
  end

  # @return [int]
  def get_position()
    send_message(find_hwnd(), WM_USER, MBC_GetPosition, 0)
  end

  # @param position [int]
  # @return [MBError]
  def set_position(position)
    send_message(find_hwnd(), WM_USER, MBC_SetPosition, position)
  end
  
  # @return [int]
  def position
    send_message(find_hwnd(), WM_USER, MBC_GetPosition, 0)
  end
  
  # @param position [int]
  def position=(position)
    send_message(find_hwnd(), WM_USER, MBC_SetPosition, position)
  end

  # Volume - Value between 0 - 100
  # @return [int]
  def get_volume()
    send_message(find_hwnd(), WM_USER, MBC_GetVolume, 0)
  end

  # Volume - Value between 0 - 100
  # @param volume [int]
  # @return [MBError]
  def set_volume(volume)
    send_message(find_hwnd(), WM_USER, MBC_SetVolume, volume)
  end
  
  # Volume - Value between 0 - 100
  # @return [int]
  def volume
    send_message(find_hwnd(), WM_USER, MBC_GetVolume, 0)
  end
  
  # Volume - Value between 0 - 100
  # @param volume [int]
  def volume=(volume)
    send_message(find_hwnd(), WM_USER, MBC_SetVolume, volume)
  end

  # Precise volume - Value between 0 - 10000
  # @return [int]
  def get_volumep()
    send_message(find_hwnd(), WM_USER, MBC_GetVolumep, 0)
  end

  # Precise volume - Value between 0 - 10000
  # @param volume [int]
  # @return [MBError]
  def set_volumep(volume)
    send_message(find_hwnd(), WM_USER, MBC_SetVolumep, volume)
  end
  
  # Precise volume - Value between 0 - 10000
  # @return [int]
  def volumep
    send_message(find_hwnd(), WM_USER, MBC_GetVolumep, 0)
  end
  
  # Precise volume - Value between 0 - 10000
  # @param volume [int]
  def volumep=(volume)
    send_message(find_hwnd(), WM_USER, MBC_SetVolumep, volume)
  end

  # Floating point volume - Value between 0.0 - 1.0
  # @return [float]
  def get_volumef()
    FloatInt.new(i:send_message(find_hwnd(), WM_USER, MBC_GetVolumef, 0))[:f]
  end

  # Floating point volume - Value between 0.0 - 1.0
  # @param volume [float]
  # @return [MBError]
  def set_volumef(volume)
    send_message(find_hwnd(), WM_USER, MBC_SetVolumef,
                 FloatInt.new(f:volume)[:i])
  end
  
  # Floating point volume - Value between 0.0 - 1.0
  # @return [float]
  def volumef
    FloatInt.new(i:send_message(find_hwnd(), WM_USER, MBC_GetVolumef, 0))[:f]
  end
  
  # Floating point volume - Value between 0.0 - 1.0
  # @param volume [float]
  def volumef=(volume)
    send_message(find_hwnd(), WM_USER, MBC_SetVolumef,
                 FloatInt.new(f:volume)[:i])
  end

  # @return [bool]
  def get_mute()
    to_b(send_message(find_hwnd(), WM_USER, MBC_GetMute, 0))
  end

  # @param mute [bool]
  # @return [MBError]
  def set_mute(mute)
    send_message(find_hwnd(), WM_USER, MBC_SetMute, to_i(mute))
  end

  # @return [bool]
  def get_shuffle()
    to_b(send_message(find_hwnd(), WM_USER, MBC_GetShuffle, 0))
  end

  # @param shuffle [bool]
  # @return [MBError]
  def set_shuffle(shuffle)
    send_message(find_hwnd(), WM_USER, MBC_SetShuffle, to_i(shuffle))
  end

  # @return [MBRepeatMode]
  def get_repeat()
    send_message(find_hwnd(), WM_USER, MBC_GetRepeat, 0)
  end

  # @param repeat [MBRepeatMode]
  # @return [MBError]
  def set_repeat(repeat)
    send_message(find_hwnd(), WM_USER, MBC_SetRepeat, repeat)
  end

  # @return [bool]
  def get_equalizer_enabled()
    to_b(send_message(find_hwnd(), WM_USER, MBC_GetEqualizerEnabled, 0))
  end

  # @param enabled [bool]
  # @return [MBError]
  def set_equalizer_enabled(enabled)
    send_message(find_hwnd(), WM_USER, MBC_SetEqualizerEnabled, to_i(enabled))
  end

  # @return [bool]
  def get_dsp_enabled()
    to_b(send_message(find_hwnd(), WM_USER, MBC_GetDspEnabled, 0))
  end

  # @param enabled [bool]
  # @return [MBError]
  def set_dsp_enabled(enabled)
    send_message(find_hwnd(), WM_USER, MBC_SetDspEnabled, to_i(enabled))
  end

  # @return [bool]
  def get_scrobble_enabled()
    to_b(send_message(find_hwnd(), WM_USER, MBC_GetScrobbleEnabled, 0))
  end

  # @param enabled [bool]
  # @return [MBError]
  def set_scrobble_enabled(enabled)
    send_message(find_hwnd(), WM_USER, MBC_SetScrobbleEnabled, to_i(enabled))
  end

  # @return [MBError]
  def show_equalizer()
    send_message(find_hwnd(), WM_USER, MBC_ShowEqualiser, 0)
  end

  # @return [bool]
  def get_auto_dj_enabled()
    to_b(send_message(find_hwnd(), WM_USER, MBC_GetAutoDjEnabled, 0))
  end

  # @return [bool]
  def get_stop_after_current_enabled()
    to_b(send_message(find_hwnd(), WM_USER, MBC_GetStopAfterCurrentEnabled, 0))
  end

  # @param enabled [bool]
  # @return [bool]
  def set_stop_after_current_enabled(enabled)
    send_message(find_hwnd(), WM_USER, MBC_SetStopAfterCurrentEnabled,
                 to_i(enabled))
  end

  # @return [bool]
  def get_crossfade()
    to_b(send_message(find_hwnd(), WM_USER, MBC_GetCrossfade, 0))
  end

  # @param crossfade [bool]
  # @return [MBError]
  def set_crossfade(crossfade)
    send_message(find_hwnd(), WM_USER, MBC_SetCrossfade, to_i(crossfade))
  end

  # @return [MBReplayGainMode]
  def get_replay_gain_mode()
    send_message(find_hwnd(), WM_USER, MBC_GetReplayGainMode, 0)
  end

  # @param mode [MBReplayGainMode]
  # @return [MBError]
  def set_replay_gain_mode(mode)
    send_message(find_hwnd(), WM_USER, MBC_SetReplayGainMode, mode)
  end

  # @param count [int]
  # @return [MBError]
  def queue_random_tracks()
    send_message(find_hwnd(), WM_USER, MBC_QueueRandomTracks, count)
  end

  # @return [int]
  def get_duration()
    send_message(find_hwnd(), WM_USER, MBC_GetDuration, 0)
  end

  # @return [String]
  def get_file_url()
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_GetFileUrl, 0)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param file_property [MBFileProperty]
  # @return [String]
  def get_file_property(file_property)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_GetFileProperty, file_property)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param field [MBMetaData]
  # @return [String]
  def get_file_tag(field)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_GetFileTag, field)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @return [String]
  def get_lyrics()
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_GetLyrics, 0)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @return [String]
  def get_downloaded_lyrics()
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_GetDownloadedLyrics, 0)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @return [String]
  def get_artwork()
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_GetArtwork, 0)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @return [String]
  def get_artwork_url()
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_GetArtworkUrl, 0)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @return [String]
  def get_downloaded_artwork()
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_GetDownloadedArtwork, 0)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @return [String]
  def get_downloaded_artwork_url()
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_GetDownloadedArtworkUrl, 0)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param fading_percent [int]
  # @return [String]
  def get_artist_picture(fading_percent)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_GetArtistPicture, fading_percent)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param local_only [bool]
  # @return [String]
  def get_artist_picture_urls(local_only)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_GetArtistPictureUrls, to_i(local_only))
    
    r = unpack_sa(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @return [String]
  def get_artist_picture_thumb()
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_GetArtistPictureThumb, 0)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @return [bool]
  def is_soundtrack()
    to_b(send_message(find_hwnd(), WM_USER, MBC_IsSoundtrack, 0))
  end

  # @param local_only [bool]
  # @return [String]
  def get_soundtrack_picture_urls(local_only)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_GetSoundtrackPictureUrls,
                      to_i(local_only))
    
    r = unpack_sa(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @return [int]
  def get_current_index()
    send_message(find_hwnd(), WM_USER, MBC_GetCurrentIndex, 0)
  end

  # @param offset [int]
  # @return [int]
  def get_next_index(offset)
    send_message(find_hwnd(), WM_USER, MBC_GetNextIndex, offset)
  end

  # @return [bool]
  def is_any_prior_tracks()
    to_b(send_message(find_hwnd(), WM_USER, MBC_IsAnyPriorTracks, 0))
  end

  # @return [bool]
  def is_any_following_tracks()
    to_b(send_message(find_hwnd(), WM_USER, MBC_IsAnyFollowingTracks, 0))
  end

  # @param fileurl [String]
  # @return [MBError]
  def play_now(fileurl)
    send_message(find_hwnd(), WM_COPYDATA, MBC_PlayNow, pack_s(fileurl))
  end

  # @param fileurl [String]
  # @return [MBError]
  def queue_next(fileurl)
    send_message(find_hwnd(), WM_COPYDATA, MBC_QueueNext, pack_s(fileurl))
  end

  # @param fileurl [String]
  # @return [MBError]
  def queue_last(fileurl)
    send_message(find_hwnd(), WM_COPYDATA, MBC_QueueLast, pack_s(fileurl))
  end
  
  # @return [MBError]
  def clear_now_playing_list()
    send_message(find_hwnd(), WM_USER, MBC_ClearNowPlayingList, 0)
  end

  # @param index [int]
  # @return [MBError]
  def remove_at(index)
    send_message(find_hwnd(), WM_USER, MBC_RemoveAt, index)
  end

  # @param from_indices [Array<int>]
  # @param to_index [int]
  # @return [MBError]
  def move_files(from_indices, to_index)
    send_message(find_hwnd(), WM_COPYDATA, MBC_MoveFiles,
                 pack_iai(from_indices, to_index))
  end
  
  # @return [MBError]
  def show_now_playing_assistant()
    send_message(find_hwnd(), WM_USER, MBC_ShowNowPlayingAssistant, 0)
  end

  # @return [bool]
  def get_show_time_remaining()
    to_b(send_message(find_hwnd(), WM_USER, MBC_GetShowTimeRemaining, 0))
  end

  # @return [bool]
  def get_show_rating_track()
    to_b(send_message(find_hwnd(), WM_USER, MBC_GetShowRatingTrack, 0))
  end

  # @return [bool]
  def get_show_rating_love()
    to_b(send_message(find_hwnd(), WM_USER, MBC_GetShowRatingLove, 0))
  end

  # @param button [MBPlayButtonType]
  # @return [bool]
  def get_button_enabled(button)
    to_b(send_message(find_hwnd(), WM_USER, MBC_GetButtonEnabled, button))
  end

  # @param index [int]
  # @return [MBError]
  def jump(index)
    send_message(find_hwnd(), WM_USER, MBC_Jump, index)
  end

  # @param query [String]
  # @param comparison [String]
  # @param fields [Array<String>]
  # @return [Array<String>]
  def search(query, comparison="Contains",
             fields=["ArtistPeople", "Title", "Album"])
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_Search,
                      pack_sssa(query, comparison, fields))
    
    r = unpack_sa(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param query [String]
  # @param comparison [String]
  # @param fields [Array<String>]
  # @return [String]
  def search_first(query, comparison="Contains",
                   fields=["ArtistPeople", "Title", "Album"])
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_SearchFirst,
                      pack_sssa(query, comparison, fields))
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param query [String]
  # @param comparison [String]
  # @param fields [Array<String>]
  # @return [Array<int>]
  def search_indices(query, comparison="Contains",
                     fields=["ArtistPeople", "Title", "Album"])
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_SearchIndices,
                      pack_sssa(query, comparison, fields))
    
    r = unpack_ia(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param query [String]
  # @param comparison [String]
  # @param fields [Array<String>]
  # @return [int]
  def search_first_index(query, comparison="Contains",
                         fields=["ArtistPeople", "Title", "Album"])
    send_message(find_hwnd(), WM_COPYDATA, MBC_SearchFirstIndex,
                 pack_sssa(query, comparison, fields))
  end

  # @param query [String]
  # @param comparison [String]
  # @param fields [Array<String>]
  # @return [MBError]
  def search_and_play_first(query, comparison="Contains",
                            fields=["ArtistPeople", "Title", "Album"])
    send_message(find_hwnd(), WM_COPYDATA, MBC_SearchAndPlayFirst,
                 pack_sssa(query, comparison, fields))
  end
  
  # @!group Now Playing List

  # @param index [int]
  # @return [String]
  def now_playing_list_get_list_file_url(index)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_NowPlayingList_GetListFileUrl, index)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param index [int]
  # @param file_property [MBFileProperty]
  # @return [String]
  def now_playing_list_get_file_property(index, file_property)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_NowPlayingList_GetFileProperty,
                      pack_i(index, file_property))
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param index [int]
  # @param field [MBMetaData]
  # @return [String]
  def now_playing_list_get_file_tag(index, field)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_NowPlayingList_GetFileTag,
                      pack_i(index, field))
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param query [String]
  # @return [MBError]
  def now_playing_list_query_files(query)
    send_message(find_hwnd(), WM_COPYDATA, MBC_NowPlayingList_QueryFiles,
                 pack_s(query))
  end

  # @return [String]
  def now_playing_list_query_get_next_file()
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_NowPlayingList_QueryGetNextFile, 0)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @return [String]
  def now_playing_list_query_get_all_files()
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_NowPlayingList_QueryGetAllFiles, 0)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param query [String]
  # @return [Array<String>]
  def now_playing_list_query_files_ex(query)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_NowPlayingList_QueryFilesEx,
                      pack_s(query))
    
    r = unpack_sa(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @return [MBError]
  def now_playing_list_play_library_shuffled()
    send_message(find_hwnd(), WM_USER,
                 MBC_NowPlayingList_PlayLibraryShuffled, 0)
  end

  # @return [int]
  def now_playing_list_get_item_count()
    send_message(find_hwnd(), WM_USER, MBC_NowPlayingList_GetItemCount, 0)
  end
  
  # @!group Playlist

  # @param playlist_url [String]
  # @return [String]
  def playlist_get_name(playlist_url)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_Playlist_GetName,
                      pack_s(playlist_url))
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param playlist_url [String]
  # @return [MBPlaylistFormat]
  def playlist_get_type(playlist_url)
    send_message(find_hwnd(), WM_COPYDATA, MBC_Playlist_GetType,
                 pack_s(playlist_url))
  end

  # @param playlist_url [String]
  # @param filename [String]
  # @return [bool]
  def playlist_is_in_list(playlist_url, filename)
    send_message(find_hwnd(), WM_COPYDATA, MBC_Playlist_IsInList,
                 pack_s(playlist_url, filename))
  end

  # @return [MBError]
  def playlist_query_playlists()
    send_message(find_hwnd(), WM_USER, MBC_Playlist_QueryPlaylists, 0)
  end

  # @return [String]
  def playlist_query_get_next_playlist()
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_Playlist_QueryGetNextPlaylist, 0)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param playlist_url [String]
  # @return [MBError]
  def playlist_query_files(playlist_url)
    send_message(find_hwnd(), WM_COPYDATA, MBC_Playlist_QueryFiles,
                 pack_s(playlist_url))
  end

  # @return [String]
  def playlist_query_get_next_file()
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_Playlist_QueryGetNextFile, 0)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @return [String]
  def playlist_query_get_all_files()
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_Playlist_QueryGetAllFiles, 0)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param playlist_url [String]
  # @return [Array<String>]
  def playlist_query_files_ex(playlist_url)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_Playlist_QueryFilesEx,
                      pack_s(playlist_url))
    
    r = unpack_sa(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param folder_name [String]
  # @param playlist_name [String]
  # @param filenames [Array<String>]
  # @return [String]
  def playlist_create_playlist(folder_name, playlist_name, filenames)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_Playlist_CreatePlaylist,
                      pack_sssa(folder_name, playlist_name, filenames))
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param playlist_url [String]
  # @return [MBError]
  def playlist_delete_playlist(playlist_url)
    send_message(find_hwnd(), WM_COPYDATA, MBC_Playlist_DeletePlaylist,
                 pack_s(playlist_url))
  end

  # @param playlist_url [String]
  # @param filenames [Array<String>]
  # @return [MBError]
  def playlist_set_files(playlist_url, filenames)
    send_message(find_hwnd(), WM_COPYDATA, MBC_Playlist_SetFiles,
                 pack_ssa(playlist_url, filenames))
  end

  # @param playlist_url [String]
  # @param filenames [Array<String>]
  # @return [MBError]
  def playlist_append_files(playlist_url, filenames)
    send_message(find_hwnd(), WM_COPYDATA, MBC_Playlist_AppendFiles,
                 pack_ssa(playlist_url, filenames))
  end

  # @param playlist_url [String]
  # @param index [int]
  # @return [MBError]
  def playlist_remove_at(playlist_url, index)
    send_message(find_hwnd(), WM_COPYDATA, MBC_Playlist_RemoveAt,
                 pack_si(playlist_url, index))
  end

  # @param playlist_url [String]
  # @param from_indices [Array<int>]
  # @param to_index [int]
  # @return [MBError]
  def playlist_move_files(playlist_url, from_indices, to_index)
    send_message(find_hwnd(), WM_COPYDATA, MBC_Playlist_MoveFiles,
                 pack_siai(playlist_url, from_indices, to_index))
  end

  # @param playlist_url [String]
  # @return [MBError]
  def playlist_play_now(playlist_url)
    send_message(find_hwnd(), WM_COPYDATA, MBC_Playlist_PlayNow,
                 pack_s(playlist_url))
  end

  # @param playlist_url [String]
  # @return [int]
  def playlist_get_item_count(playlist_url)
    send_message(find_hwnd(), WM_COPYDATA, MBC_Playlist_GetItemCount,
                 pack_s(playlist_url))
  end
  
  # @!group Library

  # @param file_url [String]
  # @param file_property [MBFileProperty]
  # @return [String]
  def library_get_file_property(file_url, file_property)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_Library_GetFileProperty,
                      pack_si(file_url, file_property))
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param file_url [String]
  # @param field [MBMetaData]
  # @return [String]
  def library_get_file_tag(file_url, field)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_Library_GetFileTag,
                      pack_si(file_url, field))
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param file_url [String]
  # @param field [MBMetaData]
  # @param value [String]
  # @return [MBError]
  def library_set_file_tag(file_url, field, value)
    send_message(find_hwnd(), WM_COPYDATA, MBC_Library_SetFileTag,
                 pack_sis(file_url, field, value))
  end

  # @param file_url [String]
  # @return [MBError]
  def library_commit_tags_to_file(file_url)
    send_message(find_hwnd(), WM_COPYDATA, MBC_Library_CommitTagsToFile,
                 pack_s(file_url))
  end

  # @param file_url [String]
  # @param lyrics_type [MBLyricsType]
  # @return [String]
  def library_get_lyrics(file_url, lyrics_type)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_Library_GetLyrics,
                      pack_si(file_url, lyrics_type))
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param file_url [String]
  # @param index [int]
  # @return [String]
  def library_get_artwork(file_url, index)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_Library_GetArtwork,
                      pack_si(file_url, index))
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param file_url [String]
  # @param index [int]
  # @return [String]
  def library_get_artwork_url(file_url, index)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_Library_GetArtworkUrl,
                      pack_si(file_url, index))
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param artist_name [String]
  # @param fading_percent [int]
  # @param fading_color [int]
  # @return [String]
  def library_get_artist_picture(artist_name, fading_percent, fading_color)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_Library_GetArtistPicture,
                      pack_si(artist_name, fading_percent, fading_color))
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param artist_name [String]
  # @param local_only [int]
  # @return [String]
  def library_get_artist_picture_urls(artist_name, local_only)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_Library_GetArtistPictureUrls,
                      pack_sb(artist_name, local_only))
    
    r = unpack_sa(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param artist_name [String]
  # @return [String]
  def library_get_artist_picture_thumb(artist_name)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_Library_GetArtistPictureThumb,
                      pack_s(artist_name))
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param file_url [String]
  # @param category [MBLibraryCategory]
  # @return [String]
  def library_add_file_to_library(file_url, category)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_Library_AddFileToLibrary,
                      pack_si(file_url, category))
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param query [String]
  # @return [MBError]
  def library_query_files(query)
    send_message(find_hwnd(), WM_COPYDATA, MBC_Library_QueryFiles,
                 pack_s(query))
  end

  # @return [String]
  def library_query_get_next_file()
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_Library_QueryGetNextFile, 0)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @return [String]
  def library_query_get_all_files()
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_Library_QueryGetAllFiles, 0)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param query [String]
  # @return [Array<String>]
  def library_query_files_ex(query)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_Library_QueryFilesEx,
                      pack_s(query))
    
    r = unpack_sa(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param artist_name [String]
  # @param minimum_artist_similarity_rating [double]
  # @return [MBError]
  def library_query_similar_artists(artist_name, minimum_artist_similarity_rating)
    send_message(find_hwnd(), WM_COPYDATA, MBC_Library_QueryLookupTable,
                 pack_sd(artist_name, minimum_artist_similarity_rating))
  end

  # @param key_tags [String]
  # @param value_tags [String]
  # @param query [String]
  # @return [MBError]
  def library_query_lookup_table(key_tags, value_tags, query)
    send_message(find_hwnd(), WM_COPYDATA, MBC_Library_QueryLookupTable,
                 pack_s(key_tags, value_tags, query))
  end

  # @param key [String]
  # @return [Array<String>]
  def library_query_get_lookup_table_value(key)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_Library_QueryGetLookupTableValue,
                      pack_s(key))
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param index [int]
  # @return [MBError]
  def library_jump(index)
    send_message(find_hwnd(), WM_USER, MBC_Library_Jump, index)
  end

  # @param query [String]
  # @param comparison [String]
  # @param fields [Array<String>]
  # @return [Array<String>]
  def library_search(query, comparison="Contains",
                     fields=["ArtistPeople", "Title", "Album"])
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_Library_Search,
                      pack_sssa(query, comparison, fields))
    
    r = unpack_sa(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param query [String]
  # @param comparison [String]
  # @param fields [Array<String>]
  # @return [String]
  def library_search_first(query, comparison="Contains",
                           fields=["ArtistPeople", "Title", "Album"])
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_Library_SearchFirst,
                      pack_sssa(query, comparison, fields))
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param query [String]
  # @param comparison [String]
  # @param fields [Array<String>]
  # @return [Array<int>]
  def library_search_indices(query, comparison="Contains",
                             fields=["ArtistPeople", "Title", "Album"])
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_Library_SearchIndices,
                      pack_sssa(query, comparison, fields))
    
    r = unpack_ia(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param query [String]
  # @param comparison [String]
  # @param fields [Array<String>]
  # @return [int]
  def library_search_first_index(query, comparison="Contains",
                                 fields=["ArtistPeople", "Title", "Album"])
    send_message(find_hwnd(), WM_COPYDATA, MBC_Library_SearchFirstIndex,
                 pack_sssa(query, comparison, fields))
  end

  # @param query [String]
  # @param comparison [String]
  # @param fields [Array<String>]
  # @return [MBError]
  def library_search_and_play_first(query, comparison="Contains",
                                    fields=["ArtistPeople", "Title", "Album"])
    send_message(find_hwnd(), WM_COPYDATA, MBC_Library_SearchAndPlayFirst,
                 pack_sssa(query, comparison, fields))
  end
  
  # @!group Setting

  # @param field [MBMetaData]
  # @return [String]
  def setting_get_field_name(field)
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_COPYDATA, MBC_Setting_GetFieldName, field)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @param field [MBMetaData]
  # @return [MBDataType]
  def setting_get_data_type(field)
    send_message(find_hwnd(), WM_USER, MBC_Setting_GetDataType, field)
  end
  
  # @!group Window

  # @return [HWND (long)]
  def window_get_handle()
    send_message(find_hwnd(), WM_USER, MBC_Window_GetHandle, 0)
  end

  # @return [HWND (pointer)]
  def window_get_handle_ptr()
    send_message_ptr(find_hwnd(), WM_USER, MBC_Window_GetHandle, 0)
  end

  # @return [MBError]
  def window_close()
    send_message(find_hwnd(), WM_USER, MBC_Window_Close, 0)
  end

  # @return [MBError]
  def window_restore()
    send_message(find_hwnd(), WM_USER, MBC_Window_Restore, 0)
  end

  # @return [MBError]
  def window_minimize()
    send_message(find_hwnd(), WM_USER, MBC_Window_Minimize, 0)
  end

  # @return [MBError]
  def window_maximize()
    send_message(find_hwnd(), WM_USER, MBC_Window_Maximize, 0)
  end

  # @param x [int]
  # @param y [int]
  # @return [MBError]
  def window_move()
    send_message(find_hwnd(), WM_USER, MBC_Window_Move, pack_i(x, y))
  end

  # @param w [int]
  # @param h [int]
  # @return [MBError]
  def window_resize()
    send_message(find_hwnd(), WM_USER, MBC_Window_Resize, pack_i(w, h))
  end

  # @return [MBError]
  def window_bring_to_front()
    send_message(find_hwnd(), WM_USER, MBC_Window_BringToFront, 0)
  end

  # @return [Array<int>]
  def window_get_position()
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_Window_GetPosition, 0)
    
    r = unpack_ii(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @return [Array<int>]
  def window_get_size()
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_Window_GetSize, 0)
    
    r = unpack_ii(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end
  
  # @!group General

  # @return [MBMusicBeeVersion]
  def get_music_bee_version()
    send_message(find_hwnd(), WM_USER, MBC_MusicBeeVersion, 0)
  end

  # @return [String]
  def get_music_bee_version_str()
    case send_message(find_hwnd(), WM_USER, MBC_MusicBeeVersion, 0)
    when MBMBV_v2_0
      "2.0"
    when MBMBV_v2_1
      "2.1"
    when MBMBV_v2_2
      "2.2"
    when MBMBV_v2_3
      "2.3"
    else
      "Unknown"
    end
  end

  # @return [String]
  def get_music_bee_version_sym()
    case send_message(find_hwnd(), WM_USER, MBC_MusicBeeVersion, 0)
    when MBMBV_v2_0
      :v2_0
    when MBMBV_v2_1
      :v2_1
    when MBMBV_v2_2
      :v2_2
    when MBMBV_v2_3
      :v2_3
    else
      :Unknown
    end
  end

  # @return [String]
  def get_plugin_version_str()
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_PluginVersion, 0)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r
  end

  # @return [Array<int>]
  def get_plugin_version()
    hwnd = find_hwnd()
    
    lr = send_message(hwnd, WM_USER, MBC_PluginVersion, 0)
    
    r = unpack_s(lr)
    
    send_message(hwnd, WM_USER, MBC_FreeLRESULT, lr)
    
    r.split(".").map { |s| s.to_i }
  end

  # @return [MBError]
  def test()
    send_message(find_hwnd(), WM_USER, MBC_Test, 0)
  end

  # @param text [String]
  # @param caption [String]
  # @return [MBError]
  def message_box(text, caption)
    send_message(find_hwnd(), WM_COPYDATA, MBC_MessageBox,
                 pack_s(text, caption))
  end
  
  
  # Converts returned String to bytes
  # @param str [String to convert]
  # @return [String]
  def to_bytes(str)
    str.encode("UTF-16LE").force_encoding("BINARY")
  end
  
  # Converts returned String to byte array
  # @param str [String to convert]
  # @return [Array<uint8>]
  def to_byte_array(str)
    str.encode("UTF-16LE").force_encoding("BINARY").unpack("C*")
  end

  private
  
  # Window Message
  WM_USER     = 0x0400
  WM_COPYDATA = 0x004A

  # @return [HWND (pointer)]
  def find_hwnd()
    find_window(nil, "MusicBee IPC Interface\0".encode("UTF-16LE"))
  end
  
  def to_b(i)
    i == 0 ? false : true
  end
  
  def to_i(b)
    b ? 1 : 0
  end
end
